package re;

import java.io.FileWriter;
import java.io.IOException;
import java.io.PrintWriter;
import java.util.*;
import java.util.Map.Entry;



/**
 * A class to represent an NFA.  You should create an NFA from a
 * regular expression using this class.  States are represented
 * as numbers (with 0 always being the start state).  Transitions
 * can be labelled with 'a' - 'z', or '@' to indicate an epsilon
 * transition.
 * 
 * The constructor for NFA creates an automaton containing only
 * state 0 and no transitions.  The other NFA methods permit you
 * to add new states and edges to the NFA, as well as to print out
 * an NFA in a form suitable for the NFASimulator (or for viewing with
 * dot).  
 */
public class NFA {

	private int acceptState = -1;

	private Vector<Hashtable<Character,Vector<Integer>>> states =
		new Vector<Hashtable<Character,Vector<Integer>>>();

	/**
	 * Create a new NFA with state 0.
	 */
	public NFA() {
		states.add(new Hashtable<Character,Vector<Integer>>());
	}

	private Hashtable<Character,Vector<Integer>> tableForState(int i) {
		Util.assertTrue(0 <= i && i < states.size(), "Bad State: " + i);
		return states.get(i);
	}

	/**
	 * Set the accepting state for the NFA.  Must be a previously allocated
	 * state.
	 */
	public void setAcceptState(int state) {
		Util.assertTrue(0 <= state && state < states.size(), "Bad State: " + state);
		acceptState = state;
	}

	/**
	 * Add a transition to the NFA.
	 * 
	 * @param start  must be a valid state in the NFA
	 * @param end  must be a valid state in the NFA
	 * @param c  the character on the edge.  <tt>'a' <= c <= 'z', or c == '@' for epsilon</tt>
	 */
	public void addTransition(int start, int end, char c) {
		Util.assertTrue(0 <= start && start < states.size(), "Bad State: " + start);
		Util.assertTrue(0 <= end && end < states.size(), "Bad State: " + end);
		Util.assertTrue('a' <= c && c <= 'z' || c == '@', "Bad symbol: '" + c + "'");
		Hashtable<Character,Vector<Integer>> t = tableForState(start);
		Vector<Integer> v = t.get(c);
		if (v == null) {
			v = new Vector<Integer>();
			t.put(c,v);
		}
		v.add(end);
	}

	/**
	 * Add a new state to the NFA and return its id number.
	 */
	public int newState() {
		int n = states.size();
		states.add(new Hashtable<Character,Vector<Integer>>());
		return n;
	}

	/**
	 * Return a String representing the NFA in a way that can be
	 * read in by <tt>nfa.NFASimulator</tt>. 
	 */
	public String toString() {
		String result = "";
		result += states.size() + "\n";
		result += acceptState + "\n";
		for (int i = 0; i < states.size(); i++) {
			Hashtable<Character,Vector<Integer>> t = states.get(i);
			result += i + " ";
			for (Entry<Character,Vector<Integer>> e : t.entrySet()) {
				result += (e.getKey() + ":(");
				boolean b = false;
				for (int j : e.getValue()) {
					if (b) { result += (","); }
					b = true;
					result += "" + j;
				}
				result += ") ";
			}
			result += ";\n";
		}
		return result;
	}

	/**
	 * Creates a file <tt>nfa.dot</tt> containing a graphical description
	 * of the NFA.  That file can then be processed with the <tt>dot</tt> utility.
	 */
	public void printDot() {
		String result = "digraph G {\n    rankdir=LR;\n";
		for (int i = 0; i < states.size(); i++) {
			if (acceptState == i) {
				result += "  S" + i + "[label=\"" + i + "\", peripheries=2,regular=true];\n";
			} else { 
				result += "  S" + i + "[label=\"" + i + "\",regular=true];\n";
			}

		}

		for (int i = 0; i < states.size(); i++) {
			Hashtable<Character,Vector<Integer>> t = states.get(i);
			for (Entry<Character,Vector<Integer>> e : t.entrySet()) {
				char c = e.getKey();
				for (int j : e.getValue()) {
					result += "  S" + i + " -> S" + j + "[label=\"" + c + "\"]" + ";\n";
				}
			}
		}
		result += "}\n";
		try {
			PrintWriter p;
			p = new PrintWriter(new FileWriter("nfa.dot"));
			p.println(result);
			p.close();
		} catch (IOException e) {
			System.out.println(e);
		}
	}
}
